@@ -28,15 +28,16 @@
 import logging
 import sys
 import warnings
-from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
+from typing import Dict, Iterable, Optional, Tuple, Union
 
 import numpy as np
 import pandas as pd
-from equilibrator_cache import Compound
+from equilibrator_cache import Compound, Reaction
 from scipy import stats
 from slugify import slugify
 
-from equilibrator_api import Q_, ComponentContribution, R, Reaction, default_T
+from equilibrator_api import Q_, ComponentContribution, R, default_T
+from equilibrator_api.phased_reaction import PhasedReaction
 
 from . import Bounds, SBtabDocument, SBtabError, SBtabTable, open_sbtabdoc
 
@@ -52,20 +53,26 @@ class StoichiometricModel(object):
 
     def __init__(
         self,
-        reactions: List[Reaction],
+        S: pd.DataFrame,
+        compound_dict: Dict[str, Compound],
+        reaction_dict: Dict[str, Reaction],
         comp_contrib: Optional[ComponentContribution] = None,
         standard_dg_primes: Optional[Q_] = None,
         dg_sigma: Optional[Q_] = None,
         bounds: Optional[Bounds] = None,
         config_dict: Optional[Dict[str, str]] = None,
-        compound_id_mapping: Callable[[Compound], str] = None,
     ) -> None:
         """Initialize a StoichiometricModel object.
 
         Parameters
         ----------
-        reactions : List[Reaction]
-            a list of Reaction objects
+        S : DataFrame
+            stoichiometric matrix, where rows are the compound IDs and columns
+            are the reaction IDs
+        compound_dict : Dict[str, Compound]
+            a dictionary of Compound objects, where the keys are their IDs
+        reaction_dict : Dict[str, Reaction]
+            a dictionary of Reaction objects, where the keys are their IDs
         comp_contrib : ComponentContribution
             a ComponentContribution object
         standard_dg_primes : Quantity, optional
@@ -77,37 +84,39 @@ class StoichiometricModel(object):
             "data/cofactors.csv" file in `equilibrator-api`)
         config_dict : dict, optional
             configuration parameters for Pathway analysis
-        compound_id_mapping : callable, optional
-            a function mapping compounds to their names in the model
         """
         self.comp_contrib = comp_contrib or ComponentContribution()
         self.water = self.comp_contrib.ccache.water
         self.is_proton = self.comp_contrib.ccache.is_proton
         self.is_water = self.comp_contrib.ccache.is_water
-        self.reactions = reactions
+        self.reaction_dict = reaction_dict
+        self.compound_dict = compound_dict
 
         if bounds is None:
-            self._bounds = Bounds.get_default_bounds(self.comp_contrib).copy()
+            self._bounds = Bounds.get_default_bounds(self.comp_contrib)
         else:
             self._bounds = bounds.copy()
 
         self.config_dict = config_dict or dict()
         self.configure()
 
-        # make sure the bounds on Water are (1, 1)
-        self._bounds.set_bounds(self.water, Q_(1.0, "M"), Q_(1.0, "M"))
-
-        self.S = self.comp_contrib.create_stoichiometric_matrix(reactions)
+        self.S = S
 
         # we keep the index to H2O handy, since we often need to remove
         # water from the list of compounds
-        self.idx_water = self.S.index.tolist().index(self.water)
+        compounds = list(self.compounds)
+        self.idx_water = compounds.index(self.water)
+        # make sure the bounds on Water are (1, 1)
+        self._bounds.set_bounds(compounds[self.idx_water], Q_(1.0, "M"), Q_(1.0, "M"))
+
+        # TODO: properly deal with H+ (like water,
+        # they should have no effect on the results)
+
         self.Nc, self.Nr = self.S.shape
 
         if standard_dg_primes is None:
             assert dg_sigma is None, (
-                "If standard_dg_primes are not "
-                "provided, dg_sigma must also be None"
+                "If standard_dg_primes are not " "provided, dg_sigma must also be None"
             )
             self.update_standard_dgs()
         else:
@@ -136,10 +145,6 @@ class StoichiometricModel(object):
             self.standard_dg_primes = standard_dg_primes
             self.dg_sigma = dg_sigma
 
-        # Set the compound names, these names can
-        # be changed later using the same function (set_compound_ids)
-        self.set_compound_ids(compound_id_mapping)
-
     def configure(self) -> None:
         """Configure the Component Contribution aqueous conditions."""
         if "p_h" in self.config_dict:
@@ -149,9 +154,7 @@ class StoichiometricModel(object):
             self.comp_contrib.p_mg = Q_(self.config_dict["p_mg"])
 
         if "ionic_strength" in self.config_dict:
-            self.comp_contrib.ionic_strength = Q_(
-                self.config_dict["ionic_strength"]
-            )
+            self.comp_contrib.ionic_strength = Q_(self.config_dict["ionic_strength"])
 
         if "temperature" in self.config_dict:
             self.comp_contrib.temperature = Q_(self.config_dict["temperature"])
@@ -172,12 +175,96 @@ class StoichiometricModel(object):
         # single metabolite indipendently). If you don't want to allow any
         # change to the mean standard_dg estimates, set ln_conc_confidence to 0.
         if "ln_conc_confidence" in self.config_dict:
-            self.ln_conc_confidence = float(
-                self.config_dict["ln_conc_confidence"]
-            )
+            self.ln_conc_confidence = float(self.config_dict["ln_conc_confidence"])
         else:
             self.ln_conc_confidence = 0.95
 
+    @property
+    def compound_ids(self) -> Iterable[str]:
+        """Get the list of compound IDs."""
+        return self.S.index
+
+    @property
+    def compounds(self) -> Iterable[Compound]:
+        """Get the list of Compound objects."""
+        return map(self.compound_dict.get, self.compound_ids)
+
+    @property
+    def compound_df(self) -> pd.DataFrame:
+        """Get a DataFrame with all the compound data.
+
+        The columns are:
+            compound_id
+            lower_bound
+            upper_bound
+
+        """
+        lbs, ubs = self.bounds
+        return pd.DataFrame(
+            data=list(
+                zip(
+                    self.compound_ids,
+                    lbs,
+                    ubs,
+                )
+            ),
+            columns=[
+                "compound_id",
+                "lower_bound",
+                "upper_bound",
+            ],
+        )
+
+    @property
+    def reaction_ids(self) -> Iterable[str]:
+        """Get the list of reaction IDs."""
+        return self.S.columns
+
+    @property
+    def reactions(self) -> Iterable[Reaction]:
+        """Get the list of Reaction objects."""
+        return map(self.reaction_dict.get, self.reaction_ids)
+
+    @property
+    def reaction_formulas(self) -> Iterable[str]:
+        """Iterate through all the reaction formulas.
+
+        :return: the reaction formulas
+        """
+        reaction_formulas = []
+        for rid in self.reaction_ids:
+            r_series = self.S[rid]
+            sparse = {
+                cid: r_series[cid] for cid in self.compound_ids if r_series[cid] != 0
+            }
+            reaction_formulas.append(StoichiometricModel._sparse_to_formula(sparse))
+        return reaction_formulas
+
+    @property
+    def reaction_df(self) -> pd.DataFrame:
+        """Get a DataFrame with all the reaction data.
+
+        The columns are:
+            reaction_id
+            reaction_formula
+            standard_dg_prime
+
+        """
+        return pd.DataFrame(
+            data=list(
+                zip(
+                    self.reaction_ids,
+                    self.reaction_formulas,
+                    self.standard_dg_primes,
+                )
+            ),
+            columns=[
+                "reaction_id",
+                "reaction_formula",
+                "standard_dg_prime",
+            ],
+        )
+
     def update_standard_dgs(self) -> None:
         """Calculate the standard ΔG' values and uncertainties.
 
@@ -187,30 +274,12 @@ class StoichiometricModel(object):
             self.standard_dg_primes,
             self.dg_sigma,
         ) = self.comp_contrib.standard_dg_prime_multi(
-            self.reactions, uncertainty_representation="fullrank"
+            self.reactions,
+            uncertainty_representation="fullrank",
         )
 
-    def set_compound_ids(
-        self, mapping: Callable[[Compound], str] = None
-    ) -> None:
-        """Use alternative compound names for outputs such as plots.
-
-        Parameters
-        ----------
-        mapping : callable, optional
-            a function mapping compounds to their names in the model
-
-        """
-        if mapping is None:
-            self.compound_ids = [
-                slugify(cpd.get_common_name(), separator="_", lowercase=False)
-                for cpd in self.S.index
-            ]
-        else:
-            self.compound_ids = list(map(mapping, self.S.index))
-
     def set_bounds(
-        self, compound_id: str, lb: Optional[Q_] = None, ub: Optional[Q_] = None
+        self, cid: str, lb: Optional[Q_] = None, ub: Optional[Q_] = None
     ) -> None:
         """Set the lower and upper bound of a compound.
 
@@ -224,25 +293,16 @@ class StoichiometricModel(object):
             the new concentration upper bound (ignored if the value is None)
         """
 
-        try:
-            cpd_idx = self.compound_ids.index(compound_id)
-        except ValueError:
-            raise KeyError(f"There is no compound with the ID {compound_id}")
+        old_ld, old_ub = self.get_bounds(cid)
+        lb = lb or old_ld
+        ub = ub or old_ub
 
-        cpd = self.S.index[cpd_idx]
-        lb = lb or self._bounds.get_lower_bound(cpd)
-        ub = ub or self._bounds.get_upper_bound(cpd)
+        assert lb.check("[concentration]"), "Bounds must be in units of concentration"
+        assert ub.check("[concentration]"), "Bounds must be in units of concentration"
 
-        assert lb.check(
-            "[concentration]"
-        ), "Bounds must be in units of concentration"
-        assert ub.check(
-            "[concentration]"
-        ), "Bounds must be in units of concentration"
+        self._bounds.set_bounds(self.compound_dict[cid], lb, ub)
 
-        self._bounds.set_bounds(cpd, lb, ub)
-
-    def get_bounds(self, compound_id: str) -> Tuple[Q_, Q_]:
+    def get_bounds(self, cid: str) -> Tuple[Q_, Q_]:
         """Get the lower and upper bound of a compound.
 
         Parameters
@@ -258,14 +318,11 @@ class StoichiometricModel(object):
             the new concentration upper bound (ignored if the value is None)
         """
 
-        try:
-            cpd_idx = self.compound_ids.index(compound_id)
-        except ValueError:
-            raise KeyError(f"There is no compound with the ID {compound_id}")
+        if cid not in self.compound_dict:
+            raise KeyError(f"There is no compound with the ID {cid}")
 
-        cpd = self.S.index[cpd_idx]
-        lb = self._bounds.get_lower_bound(cpd)
-        ub = self._bounds.get_upper_bound(cpd)
+        lb = self._bounds.get_lower_bound(self.compound_dict[cid])
+        ub = self._bounds.get_upper_bound(self.compound_dict[cid])
         return lb, ub
 
     @property
@@ -278,7 +335,12 @@ class StoichiometricModel(object):
         -------
         tuple of (lower bounds, upper bounds)
         """
-        return self._bounds.get_bounds(self.S.index)
+        return self._bounds.get_bounds(self.compounds)
+
+    @property
+    def bound_df(self) -> pd.DataFrame:
+        """Get a DataFrame with all the bounds data."""
+        return self._bounds.to_data_frame()
 
     @property
     def ln_conc_lb(self) -> np.array:
@@ -291,7 +353,7 @@ class StoichiometricModel(object):
         a NumPy array of the log lower bounds
         """
         return np.array(
-            list(map(float, self._bounds.get_ln_lower_bounds(self.S.index)))
+            list(map(float, self._bounds.get_ln_lower_bounds(self.compounds)))
         )
 
     @property
@@ -305,7 +367,7 @@ class StoichiometricModel(object):
         a NumPy array of the log upper bounds
         """
         return np.array(
-            list(map(float, self._bounds.get_ln_upper_bounds(self.S.index)))
+            list(map(float, self._bounds.get_ln_upper_bounds(self.compounds)))
         )
 
     @property
@@ -332,33 +394,26 @@ class StoichiometricModel(object):
         return ci / stats.chi2.ppf(self.ln_conc_confidence, 1) ** (0.5)
 
     @staticmethod
-    def get_compounds(reactions: Iterable[Reaction]) -> List[Compound]:
-        """Get a unique list of all compounds in all reactions.
+    def _sparse_to_formula(sparse: Dict[str, float]) -> str:
+        """Write a reaction as a chemical formula."""
 
-        :param reactions: an iterator of reactions
-        :return: a list of unique compounds
-        """
-        compounds = set()
-        for r in reactions:
-            compounds.update(r.keys())
+        def write_compound_and_coeff(cid: str, coeff: float) -> str:
+            if np.abs(coeff - 1.0) < sys.float_info.epsilon:
+                return cid
+            else:
+                return "%g %s" % (coeff, cid)
 
-        return sorted(compounds)
+        l_tokens = []
+        r_tokens = []
+        for cid, coeff in sorted(sparse.items()):
+            if coeff < 0:
+                l_tokens.append(write_compound_and_coeff(cid, -coeff))
+            elif coeff > 0:
+                r_tokens.append(write_compound_and_coeff(cid, coeff))
 
-    @property
-    def reaction_ids(self) -> Iterable[str]:
-        """Iterate through all the reaction IDs.
-
-        :return: the reaction IDs
-        """
-        return map(lambda rxn: rxn.rid, self.reactions)
-
-    @property
-    def reaction_formulas(self) -> Iterable[str]:
-        """Iterate through all the reaction formulas.
-
-        :return: the reaction formulas
-        """
-        return map(self._reaction_to_formula, self.reactions)
+        l_str = " + ".join(l_tokens)
+        r_str = " + ".join(r_tokens)
+        return f"{l_str} = {r_str}"
 
     @staticmethod
     def read_thermodynamics(
@@ -401,9 +456,7 @@ class StoichiometricModel(object):
                 val = Q_(float(val), unit)
 
                 if typ == "equilibrium constant":
-                    assert val.check(
-                        ""
-                    ), "equilibrium constants must have no units"
+                    assert val.check(""), "equilibrium constants must have no units"
                     rid2dg0[rid] = -np.log(val.m_as("")) * R * temperature
                 elif typ == "reaction gibbs energy":
                     assert val.check(
@@ -412,56 +465,19 @@ class StoichiometricModel(object):
                     val.ito("kJ/mol")
                     rid2dg0[rid] = val
                 else:
-                    raise AssertionError(
-                        "unrecognized Rate Constant Type: " + typ
-                    )
+                    raise AssertionError("unrecognized Rate Constant Type: " + typ)
             except AssertionError:
                 raise ValueError(
                     "Syntax error in Parameter table, row %d - %s" % (i, row)
                 )
         return rid2dg0
 
-    def _compound_to_id(self, compound: Compound) -> str:
-        d = dict(zip(self.S.index, self.compound_ids))
-        try:
-            return d[compound]
-        except KeyError:
-            warnings.warn(
-                f"Cannot find {compound.get_common_name()} in the model"
-            )
-            return compound.get_common_name()
-
-    def _write_compound_and_coeff(
-        self, compound: Compound, coeff: float
-    ) -> str:
-        if np.abs(coeff - 1.0) < sys.float_info.epsilon:
-            return self._compound_to_id(compound)
-        else:
-            return "%g %s" % (coeff, self._compound_to_id(compound))
-
-    def _reaction_to_formula(self, reaction: Reaction) -> str:
-        l_tokens = []
-        r_tokens = []
-        for compound, coeff in sorted(reaction.sparse.items()):
-            if self.is_proton(compound):
-                continue
-            elif coeff < 0:
-                l_tokens.append(
-                    self._write_compound_and_coeff(compound, -coeff)
-                )
-            elif coeff > 0:
-                r_tokens.append(self._write_compound_and_coeff(compound, coeff))
-
-        l_str = " + ".join(l_tokens)
-        r_str = " + ".join(r_tokens)
-        return f"{l_str} {reaction.arrow} {r_str}"
-
     @staticmethod
     def _read_network_sbtab(
         filename: Union[str, SBtabDocument],
         comp_contrib: Optional[ComponentContribution] = None,
         freetext: bool = True,
-    ) -> Tuple[pd.DataFrame, ComponentContribution, Dict[str, str]]:
+    ):
         """Initialize a Pathway object using a 'network'-only SBtab.
 
         Parameters
@@ -480,9 +496,12 @@ class StoichiometricModel(object):
 
         Returns
         -------
-        reaction_df
-        comp_contrib
+        sbtabdoc
+        S
+        reaction_dict
+        compound_dict
         config_dict
+        comp_contrib
         """
 
         def formula_to_obj(rxn_formula: str) -> Reaction:
@@ -502,11 +521,15 @@ class StoichiometricModel(object):
         sbtabdoc = open_sbtabdoc(filename)
 
         reaction_df = sbtabdoc.get_sbtab_by_id("Reaction").to_data_frame()
-        reaction_df["Reaction"] = reaction_df.ReactionFormula.apply(
+        reaction_df["ReactionObject"] = reaction_df.ReactionFormula.apply(
             formula_to_obj
         )
-        for i, rxn in enumerate(reaction_df["Reaction"]):
-            rxn.rid = f"R{i:05d}"
+        if "ID" in reaction_df.columns:
+            reaction_df.index = reaction_df.ID
+        else:
+            reaction_df.index = [f"R{i:05d}" for i in range(reaction_df.shape[0])]
+        for row in reaction_df.itertuples():
+            row.ReactionObject.rid = row.Index
 
         config_dict = {
             "p_h": str(comp_contrib.p_h),
@@ -514,7 +537,28 @@ class StoichiometricModel(object):
             "ionic_strength": str(comp_contrib.ionic_strength),
             "temperature": str(comp_contrib.temperature),
         }
-        return reaction_df, comp_contrib, config_dict
+        reaction_dict = reaction_df.ReactionObject.to_dict()
+
+        S = comp_contrib.create_stoichiometric_matrix_from_reaction_objects(
+            reaction_df.ReactionObject
+        )
+        S.columns = reaction_df.index
+
+        compound_dict = {}
+        compound_ids = []
+        for cpd in S.index:
+            cid = slugify(cpd.get_common_name(), separator="_", lowercase=False)
+            compound_ids.append(cid)
+            compound_dict[cid] = cpd
+        S.index = compound_ids
+        return (
+            sbtabdoc,
+            S,
+            reaction_dict,
+            compound_dict,
+            config_dict,
+            comp_contrib,
+        )
 
     @staticmethod
     def _read_model_sbtab(
@@ -536,7 +580,13 @@ class StoichiometricModel(object):
         -------
         stoich_model: StoichiometricModel
             A StoichiometricModel object based on the configuration SBtab
-
+        S
+        compound_dict
+        reaction_didct
+        standard_dg_primes
+        dg_sigma
+        bounds
+        config_dict
         """
         sbtabdoc = open_sbtabdoc(filename)
 
@@ -583,6 +633,7 @@ class StoichiometricModel(object):
         # use equilibrator-cache to build a dictionary of compound IDs to
         # Compound objects
         compound_df.set_index("ID", inplace=True)
+        reaction_df.set_index("ID", inplace=True)
 
         if "Identifiers" not in compound_df.columns:
             # Fallback legacy option, KEGG IDs could be provided without the
@@ -600,35 +651,38 @@ class StoichiometricModel(object):
 
         comp_contrib = comp_contrib or ComponentContribution()
 
-        compounds = []
+        compound_objects = []
         for i, row in enumerate(compound_df.itertuples()):
             if row.Identifiers:
-                compounds.append(comp_contrib.get_compound(row.Identifiers))
+                cpd = comp_contrib.get_compound(row.Identifiers)
             else:
                 # generate a novel Compound object (with a unique negative ID)
                 # just to use it later in this specific pathway
-                compounds.append(Compound(id=-i))
-        compound_df["Compound"] = compounds
+                cpd = Compound(id=-i)
+            ph_cpd = PhasedReaction.to_phased_compound(cpd)
+            if "RedoxPotential" in compound_df.columns and row.RedoxPotential:
+                ph_cpd.abundance = Q_(row.RedoxPotential)
+            compound_objects.append(ph_cpd)
 
-        if pd.isnull(compound_df.Compound).any():
+        compound_df["CompoundObject"] = compound_objects
+
+        if pd.isnull(compound_df.CompoundObject).any():
             accessions_not_found = compound_df.loc[
-                pd.isnull(compound_df.Compound), "Identifiers"
+                pd.isnull(compound_df.CompoundObject), "Identifiers"
             ]
             error_msg = str(accessions_not_found.to_dict())
             raise KeyError(
-                f"Some compounds not found in equilibrator-cache: "
-                f"{error_msg}"
+                f"Some compounds not found in equilibrator-cache: " f"{error_msg}"
             )
 
         # Read the ConcentrationConstraints table
         # ---------------------------------------
         # convert compound IDs in the bounds table to Compound objects
-        id_to_compound = compound_df.Compound.to_dict()
-        bounds_df["Compound"] = bounds_df.Compound.apply(id_to_compound.get)
-        bounds_df.set_index("Compound", inplace=True)
+        bounds_df = bounds_df.join(compound_df[["CompoundObject"]], on="Compound")
+        bounds_df.set_index("CompoundObject", inplace=True)
 
-        bounds_sbtab = sbtabdoc.get_sbtab_by_id("ConcentrationConstraint")
         try:
+            bounds_sbtab = sbtabdoc.get_sbtab_by_id("ConcentrationConstraint")
             bounds_unit = bounds_sbtab.get_attribute("Unit")
             lbs = bounds_df["Min"].apply(lambda x: Q_(float(x), bounds_unit))
             ubs = bounds_df["Max"].apply(lambda x: Q_(float(x), bounds_unit))
@@ -643,52 +697,68 @@ class StoichiometricModel(object):
 
         # Read the Reaction table
         # -----------------------
-        reactions = []
-        reaction_ids = []
-        for row in reaction_df.itertuples():
-            rxn = Reaction.parse_formula(
-                id_to_compound.get, row.ReactionFormula
+        S = comp_contrib.create_stoichiometric_matrix_from_reaction_formulas(
+            reaction_df.ReactionFormula
+        )
+        S.columns = reaction_df.index
+
+        missing_ids = set(S.index).difference(compound_df.index)
+
+        assert len(missing_ids) == 0, (
+            "Some compounds used in the `Reaction` table are not present in "
+            "the `Compound` table: " + ", ".join(missing_ids)
+        )
+
+        # TODO: if H+ is one of the compounds in S, drop it
+        # TODO: if water is not one of the compounds in S, add another row
+
+        reaction_objects = []
+        for rid in reaction_df.index:
+            sparse = dict()
+            sparse_with_phases = dict()
+            for cid in S.index:
+                if S.at[cid, rid] != 0:
+                    cpd = compound_df.at[cid, "CompoundObject"]
+                    sparse[cpd.compound] = S.at[cid, rid]
+                    sparse_with_phases[cpd] = S.at[cid, rid]
+            rxn = PhasedReaction(
+                sparse=sparse,
+                arrow="=",
+                rid=rid,
+                sparse_with_phases=sparse_with_phases,
             )
-            rxn.rid = row.ID
             if not rxn.is_balanced(ignore_atoms=("H",)):
                 warnings.warn(f"Reaction {rxn.rid} is not balanced")
-            reactions.append(rxn)
-            if row.ID in reaction_ids:
-                raise KeyError(
-                    f"Reaction IDs must be unique, but you have "
-                    f"{row.ID} twice"
-                )
-            reaction_ids.append(row.ID)
+            reaction_objects.append(rxn)
+        reaction_df["ReactionObject"] = reaction_objects
 
         thermo_sbtab = sbtabdoc.get_sbtab_by_id("Thermodynamics")
         if thermo_sbtab:
-            rid2dg0 = StoichiometricModel.read_thermodynamics(
-                thermo_sbtab, config_dict
-            )
+            rid2dg0 = StoichiometricModel.read_thermodynamics(thermo_sbtab, config_dict)
 
             # make sure all reactions have a standard Gibbs energy
-            assert set(reaction_ids).issubset(rid2dg0.keys()), (
-                "Not all reactions have a equilibrium constant or gibbs energy "
-                "in the 'Thermodynamics' table."
+            missing_ids = set(S.columns).difference(rid2dg0.keys())
+            assert len(missing_ids) == 0, (
+                "Some reactions are missing a Keq or ΔG'0 "
+                "in the `Thermodynamics` table: " + ", ".join(missing_ids)
             )
             standard_dg_primes = np.array(
-                [rid2dg0[rid].m_as("kJ/mol") for rid in reaction_ids]
+                [rid2dg0[rid].m_as("kJ/mol") for rid in reaction_df.index]
             ) * Q_("1 kJ/mol")
             dg_sigma = None
         else:
             standard_dg_primes = None
             dg_sigma = None
 
-        compound_to_id = dict(map(reversed, id_to_compound.items()))
         return (
             sbtabdoc,
-            reactions,
-            reaction_ids,
+            S,
+            compound_df.CompoundObject.to_dict(),
+            reaction_df.ReactionObject.to_dict(),
             standard_dg_primes,
             dg_sigma,
             bounds,
             config_dict,
-            compound_to_id,
         )
 
     @classmethod
@@ -722,14 +792,18 @@ class StoichiometricModel(object):
             a Pathway object
         """
         (
-            reaction_df,
-            comp_contrib,
+            sbtabdoc,
+            S,
+            reaction_dict,
+            compound_dict,
             config_dict,
-        ) = StoichiometricModel._read_network_sbtab(
-            filename, comp_contrib, freetext
-        )
+            comp_contrib,
+        ) = StoichiometricModel._read_network_sbtab(filename, comp_contrib, freetext)
+
         stoich_model = StoichiometricModel(
-            reactions=reaction_df.Reaction.tolist(),
+            S=S,
+            compound_dict=compound_dict,
+            reaction_dict=reaction_dict,
             comp_contrib=comp_contrib,
             bounds=bounds,
             config_dict=config_dict,
@@ -761,23 +835,24 @@ class StoichiometricModel(object):
         """
         (
             sbtabdoc,
-            reactions,
-            reaction_ids,
+            S,
+            compound_dict,
+            reaction_dict,
             standard_dg_primes,
             dg_sigma,
             bounds,
             config_dict,
-            compound_to_id,
         ) = cls._read_model_sbtab(filename, comp_contrib)
 
         return StoichiometricModel(
-            reactions=reactions,
+            S=S,
+            compound_dict=compound_dict,
+            reaction_dict=reaction_dict,
             comp_contrib=comp_contrib,
             standard_dg_primes=standard_dg_primes,
             dg_sigma=None,
             bounds=bounds,
             config_dict=config_dict,
-            compound_id_mapping=compound_to_id.get,
         )
 
     def to_sbtab(self) -> SBtabDocument:
@@ -794,10 +869,7 @@ class StoichiometricModel(object):
         )
 
         reaction_df = pd.DataFrame(
-            data=[
-                (rxn.rid, self._reaction_to_formula(rxn))
-                for rxn in self.reactions
-            ],
+            data=list(zip(self.reaction_ids, self.reaction_formulas)),
             columns=["!ID", "!ReactionFormula"],
         )
         reaction_sbtab = SBtabTable.from_data_frame(
@@ -807,11 +879,11 @@ class StoichiometricModel(object):
         compound_df = pd.DataFrame(
             data=[
                 (
-                    self._compound_to_id(cpd),
-                    cpd.get_common_name(),
-                    cpd.get_accession(),
+                    cid,
+                    self.compound_dict[cid].get_common_name(),
+                    self.compound_dict[cid].get_accession(),
                 )
-                for cpd in self.S.index
+                for cid in self.compound_ids
             ],
             columns=["!ID", "!Name", "!Identifiers"],
         )
@@ -850,11 +922,11 @@ class StoichiometricModel(object):
             data=[
                 (
                     "concentration",
-                    self._compound_to_id(cpd),
+                    cid,
                     f"{lb:.3g}",
                     f"{ub:.3g}",
                 )
-                for cpd, lb, ub in zip(self.S.index, lbs, ubs)
+                for cid, lb, ub in zip(self.compound_ids, lbs, ubs)
             ],
             columns=["!QuantityType", "!Compound", "!Min", "!Max"],
         )
