@@ -26,14 +26,14 @@
 # THE SOFTWARE.
 
 
-import logging
+import hashlib
 from typing import Dict, List, Tuple
 
-import numpy as np
-from equilibrator_cache.reaction import Reaction
+import pandas as pd
+from equilibrator_cache import PROTON_INCHI_KEY, Compound, Reaction
 
-from . import Q_, Compound, ccache, ureg
-from .phased_compound import PhasedCompound
+from . import Q_, ureg
+from .phased_compound import PhasedCompound, Proton, RedoxCarrier
 
 
 class PhasedReaction(Reaction):
@@ -47,7 +47,7 @@ class PhasedReaction(Reaction):
         arrow: str = "<=>",
         rid: str = None,
         sparse_with_phases: Dict[PhasedCompound, float] = None,
-    ):
+    ) -> None:
         """Create a PhasedReaction object.
 
         :param sparse: a dictionary of Compounds to stoichiometric coefficients
@@ -59,15 +59,26 @@ class PhasedReaction(Reaction):
         if sparse_with_phases is not None:
             self.sparse_with_phases = sparse_with_phases
         else:
-            self.sparse_with_phases = {
-                PhasedCompound(cpd): coeff for cpd, coeff in sparse.items()
-            }
+            self.sparse_with_phases = {}
+            for cpd, coeff in sparse.items():
+                self.sparse_with_phases[self.to_phased_compound(cpd)] = coeff
 
         if self.rid is None:
             self.rid = "COCO:R%05d" % PhasedReaction.REACTION_COUNTER
             PhasedReaction.REACTION_COUNTER += 1
 
-    def clone(self):
+    @staticmethod
+    def to_phased_compound(cpd: Compound) -> PhasedCompound:
+        """Convert a Compound object to PhasedCompound."""
+        # TODO: this is a hack, find another way to identify H+ and e-
+        if cpd.inchi_key == PROTON_INCHI_KEY:
+            return Proton(cpd)
+        elif cpd.id == 804:
+            return RedoxCarrier(cpd)
+        else:
+            return PhasedCompound(cpd)
+
+    def clone(self) -> "PhasedReaction":
         """Clone this reaction object."""
         return PhasedReaction(
             self.sparse.copy(),
@@ -76,7 +87,7 @@ class PhasedReaction(Reaction):
             self.sparse_with_phases.copy(),
         )
 
-    def reverse(self):
+    def reverse(self) -> "PhasedReaction":
         """Return a PhasedReaction with the reverse reaction."""
         sparse = {cpd: -coeff for (cpd, coeff) in self.sparse.items()}
 
@@ -86,35 +97,72 @@ class PhasedReaction(Reaction):
 
         return PhasedReaction(sparse, self.arrow, self.rid, sparse_with_phases)
 
+    def get_element_data_frame(self) -> pd.DataFrame:
+        """Tabulate the elemental composition of all reactants.
+
+        Returns
+        -------
+        DataFrame
+            A data frame where the columns are the compounds and the
+            indexes are atomic elements.
+        """
+        atom_data = [
+            (compound.compound, atom, float(count))
+            for compound in self.sparse_with_phases.keys()
+            for atom, count in compound.atom_bag.items()
+        ]
+
+        # create the elemental matrix, where each column is a compound and each
+        # row is an element (or e-)
+        return (
+            pd.DataFrame(atom_data, columns=["compound", "atom", "count"])
+            .pivot("atom", "compound", "count")
+            .fillna(0.0)
+        )
+
     @staticmethod
     def _hashable_reactants(
-        sparse_with_phases: Dict[Compound, float]
+        sparse_with_phases: Dict[PhasedCompound, float],
+        reversible: bool = False,
     ) -> Tuple[Tuple[int, str, float]]:
-        """Return a unique list of number pairs representing phased reaction.
+        """Return a unique nested tuple representing this reaction.
 
-        The list fully identifies the biochemical reaction.
-        If it is equal to another reaction's string, then they have identical
-        stoichiometry.
+        Parameters
+        ----------
+        sparse_with_phases: dict
+            a dictionary containing the compounds and stoichiometry
+        reversible : bool
+            a flag indicating whether the directionality of the reaction
+            matters or not. if `True`, the same value will be returned for
+            both the forward and backward versions.
+
+        Returns
+        -------
+            a unique tuple of 3-tuples (compound_id, phase, coefficient)
 
-        :param sparse_with_phases: a dictionary whose keys are phasedcompounds
-        and values are stoichiometric coefficients
-        :return: a unique list of pairs (compound.id, coefficient, phase)
         """
-        if len(sparse_with_phases) == 0:
-            return []
 
-        # sort according to compound ID and normalize the stoichiometric
-        # coefficients such that the coeff of the reactant with the lowest
-        # ID will be 1
-        sorted_compound_list = sorted(
-            sparse_with_phases.keys(), key=lambda c: c.id
-        )
+        if len(sparse_with_phases) == 0:
+            return tuple()
+
+        # sort according to compound ID and normalize
+        sorted_compound_list = sorted(sparse_with_phases.keys(), key=lambda c: c.id)
+
+        # normalize the stoichiometric coefficients such that the highest
+        # absolute value is 1
         max_coeff = max(map(abs, sparse_with_phases.values()))
         if max_coeff == 0:
             raise Exception("All stoichiometric ceofficients are 0")
+
+        # if the reversible flag is on, make sure that the normalized
+        # coefficient of the first compound (i.e. the one with the lowest ID)
+        # is positive.
+        if reversible and sparse_with_phases[sorted_compound_list[0]] < 0:
+            max_coeff *= -1.0
+
         norm_factor = 1.0 / max_coeff
         r_list = [
-            (cpd.id, cpd.phase, norm_factor * sparse_with_phases[cpd])
+            (cpd.id, cpd.phase, float(norm_factor * sparse_with_phases[cpd]))
             for cpd in sorted_compound_list
         ]
         return tuple(r_list)
@@ -122,14 +170,49 @@ class PhasedReaction(Reaction):
     def __hash__(self) -> int:
         """Return a hash of the PhasedReaction.
 
+        Note that the standard Python library `hash()` includes randomization
+        that is initialized each time the Python kernel starts. Therefore
+        this hash will not be consistent between runs. For quick reaction
+        lookup, use `PhasedReaction.hash_md5()`.
+
+        Returns
+        -------
+        str
+            a hash of the Reaction.
+        """
+        if self.is_empty():
+            return 0
+        reactants = PhasedReaction._hashable_reactants(self.sparse_with_phases)
+        return hash(reactants)
+
+    def hash_md5(self, reversible: bool = True) -> str:
+        """Return a MD5 hash of the PhasedReaction.
+
         This hash is useful for finding reactions with the exact same
         stoichiometry. We create a unique formula string based on the
         Compound IDs and coefficients.
 
-        :return: a hash of the Reaction.
+        Parameters
+        ----------
+        reversible : bool
+            a flag indicating whether the directionality of the reaction
+            matters or not. if `True`, the same value will be returned for
+            both the forward and backward versions.
+
+        Returns
+        -------
+        hash : str
+            a unique hash string representing the Reaction.
         """
-        reactants = PhasedReaction._hashable_reactants(self.sparse_with_phases)
-        return hash(reactants)
+        if self.is_empty():
+            return ""
+        reactants = PhasedReaction._hashable_reactants(
+            self.sparse_with_phases, reversible=reversible
+        )
+
+        # convert the unique tuple to bytes, since hashlib requires it.
+        reactants_str = str(reactants).encode()
+        return hashlib.md5(reactants_str).hexdigest()
 
     def __eq__(self, other: "PhasedReaction") -> bool:
         """Check if two reactions are equal.
@@ -138,9 +221,7 @@ class PhasedReaction(Reaction):
         coefficients by a scalar.
         """
         reactants = PhasedReaction._hashable_reactants(self.sparse_with_phases)
-        other_reactants = PhasedReaction._hashable_reactants(
-            other.sparse_with_phases
-        )
+        other_reactants = PhasedReaction._hashable_reactants(other.sparse_with_phases)
         return reactants == other_reactants
 
     def __lt__(self, other: "PhasedReaction") -> bool:
@@ -149,9 +230,7 @@ class PhasedReaction(Reaction):
         Used mainly for using reactions as dictionary keys.
         """
         reactants = PhasedReaction._hashable_reactants(self.sparse_with_phases)
-        other_reactants = PhasedReaction._hashable_reactants(
-            other.sparse_with_phases
-        )
+        other_reactants = PhasedReaction._hashable_reactants(other.sparse_with_phases)
         return reactants < other_reactants
 
     def set_abundance(self, compound: Compound, abundance: ureg.Quantity):
@@ -171,9 +250,7 @@ class PhasedReaction(Reaction):
             if phased_compound.compound == compound:
                 phased_compound.phase = phase
 
-    def get_phased_compound(
-        self, compound: Compound
-    ) -> Tuple[PhasedCompound, float]:
+    def get_phased_compound(self, compound: Compound) -> Tuple[PhasedCompound, float]:
         """Get the PhasedCompound object by the Compound object."""
         # TODO: This is not ideal. We should try to not keep two dictionaries (
         #  one with Compounds and one with PhasedCompounds).
@@ -237,28 +314,33 @@ class PhasedReaction(Reaction):
 
         self.sparse_with_phases[PhasedCompound(compound)] = coeff
 
-    @ureg.check(None, None, "[concentration]", "[temperature]")
+    @ureg.check(None, None, "[concentration]", "[temperature]", "")
     def separate_stored_dg_prime(
         self,
         p_h: ureg.Quantity,
         ionic_strength: ureg.Quantity,
         temperature: ureg.Quantity,
+        p_mg: ureg.Quantity,
     ) -> Tuple[Reaction, ureg.Quantity]:
         """Split the PhasedReaction to aqueous phase and all the rest.
 
         :param p_h:
         :param ionic_strength:
         :param temperature:
+        :param p_mg:
         :return: a tuple (residual_reaction, stored_dg_prime) where
         residual_reaction is a Reaction object (excluding the compounds that
         had stored values), and stored_dg_prime is the total dG' of the
         compounds with stored values (in kJ/mol).
         """
-        stored_dg_prime = 0.0
+        stored_dg_prime = Q_(0.0, "kJ/mol")
         sparse = {}  # all compounds without stored dgf values
         for phased_compound, coeff in self.sparse_with_phases.items():
             dgf_prime = phased_compound.get_stored_standard_dgf_prime(
-                p_h=p_h, ionic_strength=ionic_strength, temperature=temperature
+                p_h=p_h,
+                ionic_strength=ionic_strength,
+                temperature=temperature,
+                p_mg=p_mg,
             )
 
             if dgf_prime is None:
@@ -266,9 +348,13 @@ class PhasedReaction(Reaction):
             else:
                 stored_dg_prime += coeff * dgf_prime
 
+        # if the reaction contains a general redox pair (represented by free
+        # electrons), translate their potential to standard_dg_prime and add
+        # it to stored_dg_prime.
+
         return Reaction(sparse, arrow=self.arrow, rid=self.rid), stored_dg_prime
 
-    def separate_stored_dg(self,) -> Tuple[Reaction, ureg.Quantity]:
+    def separate_stored_dg(self) -> Tuple[Reaction, ureg.Quantity]:
         """Split the PhasedReaction to aqueous phase and all the rest.
 
         :return: a tuple (residual_reaction, stored_dg) where
@@ -313,65 +399,6 @@ class PhasedReaction(Reaction):
             ddg_over_rt += coeff * phased_compound.ln_physiological_abundance
         return ddg_over_rt
 
-    def balance_by_oxidation(self):
-        """Convert an unbalanced reaction into oxidation.
-
-        By adding H2O, O2, Pi, CO2, and NH4+ to both sides.
-        """
-        # We need to make sure that the number of balancing compounds is the
-        # same as the number of elements we are trying to balance. Here both
-        # numbers are 6, i.e. the elements are (e-, H, O, P, C, N)
-        balancing_inchis = [
-            "InChI=1S/p+1",  # H+
-            "InChI=1S/H2O/h1H2",  # H2O
-            "InChI=1S/O2/c1-2",  # O2
-            "InChI=1S/H3O4P/c1-5(2,3)4/h(H3,1,2,3,4)/p-2",  # Pi
-            "InChI=1S/CO2/c2-1-3",  # CO2
-            "InChI=1S/H3N/h1H3/p+1",  # NH4+
-        ]
-        compounds = list(map(ccache.get_compound_by_inchi, balancing_inchis))
-
-        S = ccache.get_element_data_frame(compounds)
-        balancing_atoms = S.index
-
-        atom_bag = self._get_reaction_atom_bag()
-        if atom_bag is None:
-            logging.warning(
-                "Cannot balance this reaction due to"
-                " missing chemical formulas"
-            )
-            return self
-        atom_vector = np.array(
-            list(map(lambda a: atom_bag.get(a, 0), balancing_atoms))
-        )
-
-        other_atoms = set(atom_bag.keys()).difference(balancing_atoms)
-        if other_atoms:
-            raise ValueError(
-                f"Cannot oxidize {self} only with these atoms: "
-                f"{other_atoms}"
-            )
-
-        # solve the linear equation S * i = a,
-        # and ignore small coefficients (< 1e-3)
-        imbalance = (-np.linalg.inv(S.values) @ atom_vector).round(3)
-
-        for compound, coeff in zip(S.columns, imbalance.flat):
-            self.add_stoichiometry(compound, coeff)
-
-        return self
-
-    @classmethod
-    def get_oxidation_reaction(cls, compound: Compound):
-        """Generate an oxidation Reaction for a single compound.
-
-        Generate a Reaction object which represents the oxidation reaction
-        of this compound using O2. If there are N atoms, the product must
-        be NH3 (and not N2) to represent biological processes.
-        Other atoms other than C, N, H, and O will raise an exception.
-        """
-        return cls({compound: -1}).balance_by_oxidation()
-
     def serialize(self) -> List[dict]:
         """Return a serialized version of all the reaction thermo data."""
         list_of_compounds = []
